import List
import IO

newtype Natural = Natural Integer

-- 15 :: Natural

-- Natural 15 :: Natural

x = Natural 15 :: Natural

data Tree a = Leaf | Node { value :: a,
                            left :: Tree a,
                            right :: Tree a }

-- add x y = x s y sszege
add     :: Integer -> Integer -> Integer
add x y = x + y

max' a b | a >= b    = a
	 | otherwise = b

take' m xs = case (m,xs) of
	     (0,_)    -> []
	     (_,[])   -> []
	     (m,x:xs) -> x : take (m-1) xs

distance (x1,y1) (x2,y2) = let xd = x1-x2
			       yd = y1-y2
			   in sqrt(xd*xd + yd*yd)

ones = 1 : ones

upFrom n = n : upFrom (n+1)

squares = map (^2) (upFrom 0)

fib1 = 1 : 1 : fib1 +: (tail fib1)
    where (x:xs) +: (y:ys) = x+y : xs +: ys

fib2 @ (_:tfib2) = 1 : 1 : zipWith (+) fib2 tfib2

fib3 = 1 : 1 : [ a+b | (a,b) <- zip fib3 tfib3 ]
    where _:tfib3 = fib3

data Fraction = !Integer :/ !Integer
-- (\(x :/ y) -> x) (1 :/ undefined)

cartesian       :: [a] -> [b] -> [(a,b)]
cartesian xs ys = [ (x,y) | x <- xs, y <- ys ]

fromThenTo n n' m = nn'm
    where nn'm  = takeWhile p (n : map ((n'-n) +) nn'm)
          p | n' >= n   = (m >=)
            | otherwise = (m <=)

x `member_of` []     = False
x `member_of` (y:ys) = x == y || x `member_of` ys

instance Eq Fraction where
    (a:/b) == (x:/y) = a*y == x*b

instance Eq a => Eq (Tree a) where
    Leaf          == Leaf          = True
    Node v1 l1 r1 == Node v2 l2 r2 = (v1,l1,r1) == (v2,l2,r2)
    _             == _             = False

instance Functor Tree where
    fmap f  Leaf        = Leaf
    fmap f (Node v l r) = Node (f v) (fmap f l) (fmap f r)

showTree              :: Show a => Tree a -> String
showTree Leaf         = "<>"
showTree (Node v l r) = "<" ++ showTree l ++ "|" ++ 
			       show v     ++ "|" ++
			       showTree r ++ ">"

showsTree              :: Show a => Tree a -> String -> String
showsTree Leaf         = ("<>"++)
showsTree (Node v l r) = ('<':) . (showsTree l) . ('|':) .
			          (shows v)     . ('|':) .
				  (showsTree r) . ('>':)

instance Show a => Show (Tree a) where
    showsPrec _ = showsTree

-- data Tree a = Leaf | Node a (Tree a) (Tree a)
-- 	      deriving (Eq, Ord, Show)

data Peano = Zero | Succ Peano
	   deriving (Eq, Ord, Show)

instance Num Peano where
    Zero   + m            = m
    Succ n + m            = n + Succ m

    n      - Zero         = n
    Succ n - Succ m       = n - m
    Zero   - m            = error "Peano.(-): negative number"

    abs                   = id
    signum n              = 1

    fromInteger 0         = Zero
    fromInteger n | n > 0 = Succ (fromInteger (n-1))

instance Real Peano where
    toRational            = toRational . toInteger

instance Enum Peano where
    succ n                = Succ n

    pred  Zero            = error "Peano.pred: negative number"
    pred (Succ n)         = n

    toEnum                = fromInteger . toInteger
    fromEnum              = fromInteger . toInteger

instance Integral Peano where
    n `quotRem` Zero      = error "Peano.quotRem: division by zero"
    n `quotRem` d         = qR n d Zero n
	where qR  Zero     Zero    q r = (Succ q, Zero)
	      qR  Zero     d       q r = (q, r)
	      qR  n        Zero    q r = qR n d (Succ q) n
	      qR (Succ n) (Succ d) q r = qR n d q r

    toInteger Zero        = 0
    toInteger (Succ n)    = 1 + toInteger n

-----------------------------------------------------------------------
-----------------------------------------------------------------------
-----------------------------------------------------------------------

type Query  = String
type DB     = ()
type Record = String

doQuery       :: Query -> DB -> Maybe Record
doQuery "" () = Nothing
doQuery q  () = Just q

thenMB        :: Maybe a -> (a -> Maybe b) -> Maybe b
mB `thenMB` f = case mB of
		Nothing -> Nothing
		Just a  -> f a

type StateT s a = s -> (a,s)

thenST        :: StateT s a -> (a -> StateT s b) -> StateT s b
st `thenST` f = \s -> let (v,s') = st s
		      in f v s'

returnST   :: a -> StateT s a
returnST a = \s -> (a,s)

type MbStateT s a = s -> Maybe (a,s)

thenMST        :: MbStateT s a -> (a -> MbStateT s b) -> MbStateT s b
st `thenMST` f = \s -> case st s of
				 Nothing     -> Nothing
				 Just (v,s') -> f v s'

returnMST   :: a -> MbStateT s a
returnMST v = \s -> Just (v,s)

then_MST           :: MbStateT s a -> MbStateT s b -> MbStateT s b
st1 `then_MST` st2 = st1 `thenMST` \_ -> st2


newtype ST s a = ST (StateT s a)

instance Monad (ST s) where
    ST st >>= f = ST (\s -> let (v,s') = st s
                                ST st' = f v
		            in st' s')
    return a    = ST (\s -> (a,s))

r :: Maybe Record
r = doQuery "egy" ()            >>= \r1 -> 
    doQuery (r1 ++ " ketto") () >>= \r2 ->
    doQuery (r2 ++ " harom") ()

type ImpS = (Integer,Integer)

getX, getY :: ST ImpS Integer
getX       = ST (\(x,y) -> (x, (x,y)))
getY       = ST (\(x,y) -> (y, (x,y)))

putX, putY :: Integer -> ST ImpS ()
putX x'    = ST (\(x,y) -> ((), (x',y)))
putY y'    = ST (\(x,y) -> ((), (x,y')))

gcdST :: ST ImpS Integer
gcdST = do x <- getX
           y <- getY
	   case compare x y of
	      EQ -> return x
	      LT -> do putY (y-x)
		       gcdST
	      GT -> do putX (x-y)
		       gcdST

applyST         :: ST s a -> StateT s a
applyST (ST st) = st

-- gcd' x y = fst $ applyST gcdST (x,y)

gcd' x y = fst (applyST gcdST (x,y))

getChar' :: IO Char
getChar' =  getChar `catch` eofHandler
    where eofHandler e = if isEOFError e
			 then return '\n'
                         else ioError e
