import IO
import List
import Monad

type Query  = String
type DB     = ()
type Record = String

doQuery       :: Query -> DB -> Maybe Record
doQuery "" () = Nothing
doQuery q  () = Just q

thenMB        :: Maybe a -> (a -> Maybe b) -> Maybe b
mB `thenMB` f = case mB of
		Nothing -> Nothing
		Just a  -> f a

type StateT s a = s -> (a,s)

thenST        :: StateT s a -> (a -> StateT s b) -> StateT s b
st `thenST` f = \s -> let (v,s') = st s
		      in f v s'

returnST   :: a -> StateT s a
returnST a = \s -> (a,s)

type MbStateT s a = s -> Maybe (a,s)

thenMST        :: MbStateT s a -> (a -> MbStateT s b) -> MbStateT s b
st `thenMST` f = \s -> case st s of
				 Nothing     -> Nothing
				 Just (v,s') -> f v s'

returnMST   :: a -> MbStateT s a
returnMST v = \s -> Just (v,s)

then_MST           :: MbStateT s a -> MbStateT s b -> MbStateT s b
st1 `then_MST` st2 = st1 `thenMST` \_ -> st2


newtype ST s a = ST (StateT s a)

instance Monad (ST s) where
    ST st >>= f = ST (\s -> let (v,s') = st s
                                ST st' = f v
		            in st' s')
    return a    = ST (\s -> (a,s))

r :: Maybe Record
r = doQuery "egy" ()            >>= \r1 -> 
    doQuery (r1 ++ " ketto") () >>= \r2 ->
    doQuery (r2 ++ " harom") ()

type ImpS = (Integer,Integer)

getX, getY :: ST ImpS Integer
getX       = ST (\(x,y) -> (x, (x,y)))
getY       = ST (\(x,y) -> (y, (x,y)))

putX, putY :: Integer -> ST ImpS ()
putX x'    = ST (\(x,y) -> ((), (x',y)))
putY y'    = ST (\(x,y) -> ((), (x,y')))

gcdST :: ST ImpS Integer
gcdST = do x <- getX
           y <- getY
	   case compare x y of
	      EQ -> return x
	      LT -> do putY (y-x)
		       gcdST
	      GT -> do putX (x-y)
		       gcdST

applyST         :: ST s a -> StateT s a
applyST (ST st) = st

-- gcd' x y = fst $ applyST gcdST (x,y)
gcd' x y = fst $ applyST gcdST (x,y)

getChar' :: IO Char
getChar' =  getChar `catch` eofHandler
    where eofHandler e = if isEOFError e
			 then return '\n'
                         else ioError e

f :: Int -> Maybe [Int]
f ix = do let ls = [Just [1,2,3], Nothing, Just [], Just [7..10]]
          x:xs <- ls!!ix -- a pattern match failure calls "fail"
          return xs

x = [ (x,y) |
       x <- [1,2,3],
       y <- [1,2,3],             
       x /= y ]                          

y = do x    <- [1,2,3]
       y    <- [1,2,3]
       True <- return (x /= y)
       return (x,y)

yx = [1,2,3] >>= (\x ->
     [1,2,3] >>= (\y ->
     return (x /= y) >>= (\b ->
     case b of
     True -> return (x,y)
     False -> fail "")))

yy = [1,2,3] >>= (\x ->
     [1,2,3] >>= (\y ->
     return (x /= y) >>= (\b ->
     case b of
     True -> [(x,y)]
     False -> [])))

yz = [1,2,3] >>= (\x ->
     [1,2,3] >>= (\y ->
     [(x /= y)] >>= (\b ->
     case b of
     True -> [(x,y)]
     False -> [])))

-- msum     :: MonadPlus m => [m a] -> m a
-- msum     = foldr mplus mzero

liftM    :: Monad m => (a -> b) -> (m a -> m b)
liftM f  = \a -> do { a' <- a; return (f a') }

myGetLine     :: IO String
myGetLine     = do c <- getChar
                   if c == '\n' then return ""
                     else do cs <- myGetLine; return (c:cs)
